﻿#region File Description
//-----------------------------------------------------------------------------
// CpuSkinnedModelPart.cs
//
// Microsoft XNA Community Game Platform
// Copyright (C) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
#endregion

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace CpuSkinningDataTypes
{
    public class CpuSkinnedModelPart
    {
        private readonly int triangleCount;
        private readonly int vertexCount;
        public readonly CpuVertex[] cpuVertices;
        public readonly VertexPositionNormalTexture[] gpuVertices;

        private readonly DynamicVertexBuffer vertexBuffer;
        private readonly IndexBuffer indexBuffer;


        public DynamicVertexBuffer VertexBuffer {get{return vertexBuffer;}}
        public int VertexCount { get { return vertexCount; } }
        

        public BasicEffect Effect { get; internal set; }
        
		public IndexBuffer IndexBuffer { get { return indexBuffer; } }
        
        internal CpuSkinnedModelPart(int triangleCount, CpuVertex[] vertices, IndexBuffer indexBuffer)
        {
            this.triangleCount = triangleCount;
            this.vertexCount = vertices.Length;
            this.cpuVertices = vertices;
            this.indexBuffer = indexBuffer;
       
            // create our GPU resources
            gpuVertices = new VertexPositionNormalTexture[cpuVertices.Length];
            vertexBuffer = new DynamicVertexBuffer(indexBuffer.GraphicsDevice, typeof(VertexPositionNormalTexture), cpuVertices.Length, BufferUsage.WriteOnly);

            // copy texture coordinates once since they don't change with skinnning
            for (int i = 0; i < cpuVertices.Length; i++)
            {
                gpuVertices[i].TextureCoordinate = cpuVertices[i].TextureCoordinate;
            }
        }

        public void SetBoneTransforms(Matrix[] bones)
        {
            // skin all of the vertices
            for (int i = 0; i < vertexCount; i++)
            {
                CpuSkinningHelpers.SkinVertex(
                    bones,
                    ref cpuVertices[i].Position,
                    ref cpuVertices[i].Normal,
                    ref cpuVertices[i].BlendIndices,
                    ref cpuVertices[i].BlendWeights,
                    out gpuVertices[i].Position,
                    out gpuVertices[i].Normal);
            }

            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);
        }

        public void SetBoneTransformsNEON(Matrix[] bones)
        {
            // skin all of the vertices
            for (int i = 0; i < vertexCount; i++)
            {
                CpuSkinningHelpers.SkinVertexNEON(
                    bones,
                    ref cpuVertices[i].Position,
                    ref cpuVertices[i].Normal,
                    ref cpuVertices[i].BlendIndices,
                    ref cpuVertices[i].BlendWeights,
                    out gpuVertices[i].Position,
                    out gpuVertices[i].Normal);
            }

            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.Discard);        
        }
        
        public void SetBoneTransformsFlat(Matrix[] bones)
        {
            // skin all of the vertices
            for (int i = 0; i < vertexCount; i++)
            {
                int b0 = (int)cpuVertices[i].BlendIndices.X;
                int b1 = (int)cpuVertices[i].BlendIndices.Y;
                int b2 = (int)cpuVertices[i].BlendIndices.Z;
                int b3 = (int)cpuVertices[i].BlendIndices.W;

                float w1 = cpuVertices[i].BlendWeights.X;
                float w2 = cpuVertices[i].BlendWeights.Y;
                float w3 = cpuVertices[i].BlendWeights.Z;
                float w4 = cpuVertices[i].BlendWeights.W;

                Matrix skinnedTransformSum;
                
     
                Matrix m1 = bones[b0];
                Matrix m2 = bones[b1];
                Matrix m3 = bones[b2];
                Matrix m4 = bones[b3];
                skinnedTransformSum = new Matrix();
                skinnedTransformSum.M11 = (m1.M11 * w1) + (m2.M11 * w2) + (m3.M11 * w3) + (m4.M11 * w4);
                skinnedTransformSum.M12 = (m1.M12 * w1) + (m2.M12 * w2) + (m3.M12 * w3) + (m4.M12 * w4);
                skinnedTransformSum.M13 = (m1.M13 * w1) + (m2.M13 * w2) + (m3.M13 * w3) + (m4.M13 * w4);
                skinnedTransformSum.M21 = (m1.M21 * w1) + (m2.M21 * w2) + (m3.M21 * w3) + (m4.M21 * w4);
                skinnedTransformSum.M22 = (m1.M22 * w1) + (m2.M22 * w2) + (m3.M22 * w3) + (m4.M22 * w4);
                skinnedTransformSum.M23 = (m1.M23 * w1) + (m2.M23 * w2) + (m3.M23 * w3) + (m4.M23 * w4);
                skinnedTransformSum.M31 = (m1.M31 * w1) + (m2.M31 * w2) + (m3.M31 * w3) + (m4.M31 * w4);
                skinnedTransformSum.M32 = (m1.M32 * w1) + (m2.M32 * w2) + (m3.M32 * w3) + (m4.M32 * w4);
                skinnedTransformSum.M33 = (m1.M33 * w1) + (m2.M33 * w2) + (m3.M33 * w3) + (m4.M33 * w4);
                skinnedTransformSum.M41 = (m1.M41 * w1) + (m2.M41 * w2) + (m3.M41 * w3) + (m4.M41 * w4);
                skinnedTransformSum.M42 = (m1.M42 * w1) + (m2.M42 * w2) + (m3.M42 * w3) + (m4.M42 * w4);
                skinnedTransformSum.M43 = (m1.M43 * w1) + (m2.M43 * w2) + (m3.M43 * w3) + (m4.M43 * w4);
                skinnedTransformSum.M14 = 0.0f;
                skinnedTransformSum.M24 = 0.0f;
                skinnedTransformSum.M34 = 0.0f;
                skinnedTransformSum.M44 = 1.0f;

                // Support the 4 Bone Influences - Position then Normal
                Vector3.Transform(ref cpuVertices[i].Position, ref skinnedTransformSum, out gpuVertices[i].Position);
                Vector3.TransformNormal(ref cpuVertices[i].Normal, ref skinnedTransformSum, out gpuVertices[i].Normal);
            }

            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.NoOverwrite);
        }
        
        public void SetBoneTransformsFlatNEON(Matrix[] bones)
        {
            // skin all of the vertices
            for (int i = 0; i < vertexCount; i++)
            {
                int b0 = (int)cpuVertices[i].BlendIndices.X;
                int b1 = (int)cpuVertices[i].BlendIndices.Y;
                int b2 = (int)cpuVertices[i].BlendIndices.Z;
                int b3 = (int)cpuVertices[i].BlendIndices.W;

                float w1 = cpuVertices[i].BlendWeights.X;
                float w2 = cpuVertices[i].BlendWeights.Y;
                float w3 = cpuVertices[i].BlendWeights.Z;
                float w4 = cpuVertices[i].BlendWeights.W;

                Matrix skinnedTransformSum;

                // Moblunatic claims ~40% faster.
                // http://forums.create.msdn.com/forums/p/55123/335148.aspx
                // This is true on WP7 with SIMD enabled. 
                // On WP8/Monogame it is slower than the original code.                
                
                Matrix mm1, mm2, mm3, mm4;
                Matrix.Multiply(ref bones[b0], w1, out mm1);
                Matrix.Multiply(ref bones[b1], w2, out mm2);
                Matrix.Multiply(ref bones[b2], w3, out mm3);
                Matrix.Multiply(ref bones[b3], w4, out mm4);

                Matrix.Add(ref mm1, ref mm2, out skinnedTransformSum);
                Matrix.Add(ref skinnedTransformSum, ref mm3, out skinnedTransformSum);
                Matrix.Add(ref skinnedTransformSum, ref mm4, out skinnedTransformSum);
                skinnedTransformSum.M14 = 0.0f;
                skinnedTransformSum.M24 = 0.0f;
                skinnedTransformSum.M34 = 0.0f;
                skinnedTransformSum.M44 = 1.0f;

                // Support the 4 Bone Influences - Position then Normal
                Vector3.Transform(ref cpuVertices[i].Position, ref skinnedTransformSum, out gpuVertices[i].Position);
                Vector3.TransformNormal(ref cpuVertices[i].Normal, ref skinnedTransformSum, out gpuVertices[i].Normal);
            }

            // put the vertices into our vertex buffer
            vertexBuffer.SetData(gpuVertices, 0, vertexCount, SetDataOptions.NoOverwrite);
        }
        
        public void Draw()
        {
            GraphicsDevice graphics = Effect.GraphicsDevice;

            // set our buffers on the device
            graphics.Indices = indexBuffer;
            graphics.SetVertexBuffer(vertexBuffer);

            foreach (EffectPass pass in Effect.CurrentTechnique.Passes)
            {
                pass.Apply();
                graphics.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, vertexCount, 0, triangleCount);
            }

            graphics.Indices = null;
            graphics.SetVertexBuffer(null);
        }
    }
}
